/**
  *  \file client/mainopts.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include "client/mainopts.h"
#include "arch/platform.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/struct.h"
#include "util/console.h"
#include "util/nls.h"
#include "util/misc.h"
#include "io/dirs.h"
#include "util/recode.h"

/** Initialize. */
PCC2OptionParser::PCC2OptionParser()
    : num_nonoptions(0), player(0),
      mode(mUnset)
{
    static const Option options[] = {
        { "help",    help,       0, false,
          N_(" -help              Show help screen") },
        { "charset", setCharset, 0, true,
          N_(" -charset EXT[:INT] Set character set translation") },
        { "dir",     setDirectoryMode, 0, false,
          N_(" -dir               Start with directory selection") },
    };
    addOptionSet(options, countof(options));
}

/** Handle Non-Options. The first non-option is the game directory
    name, the second (optional) is the root directory. Player number
    can stand anywhere. */
void
PCC2OptionParser::handleNonoption(const string_t& str)
{
    int x;
    if (player == 0 && stringValue(str, x) && x > 0 && x <= NUM_PLAYERS) {
        player = x;
    } else {
        if (num_nonoptions >= MAX_NON_OPTIONS) {
            handleError(_("Too many parameters specified."));
        } else {
            nonopts[num_nonoptions++] = str;
        }
    }
}

/** Handle "-help" option: show help screen. */
void
PCC2OptionParser::help(StandardCommandLineParser& par,
                       const StandardCommandLineParser::Option&)
{
    console << _("Usage: pcc-v2 [-options] gamedir [playerid] [rootdir]\n"
                 "\n"
                 "Valid options:\n");
    par.printHelp(console);
    console << _("\n(c) 2001-2011 Stefan Reuther <Streu@gmx.de>\n");
    Platform::showOutput();
    std::exit(0);
}

/** Handle "-charset" option: set character set translation. */
void
PCC2OptionParser::setCharset(StandardCommandLineParser& par,
                             const StandardCommandLineParser::Option&)
{
    string_t csName = par.getArg();
    CharacterSet cs = CharacterSet::getCharsetByName(csName);
    if (!cs.isValid())
        console.write(LOG_WARN, format(_("Character set `%s' is unknown"), csName));
    setGameCharacterSet(cs);
}

/** Handle "-dir" option. */
void
PCC2OptionParser::setDirectoryMode(StandardCommandLineParser& par,
                                   const StandardCommandLineParser::Option&)
{
    static_cast<PCC2OptionParser&>(par).mode = mChooseGame;
}

/** Error handler. Instead of throwing an exception, show a message
    and bail out. */
void
PCC2OptionParser::handleError(const string_t& str)
{
    console.write(LOG_ERROR, str);
    console.write(LOG_ERROR, _("Try `pcc-v2 --help' for help."));
    Platform::showOutput();
    std::exit(1);
}

/** Parsing finished. Postprocess input. */
void
PCC2OptionParser::parseFinished()
{
    /* Set directory defaults */
    if (getNumNonoptions() < 2)
        nonopts[1] = Platform::getDataDirectory();
    if (getNumNonoptions() < 1)
        nonopts[0] = Directory::getWorkingDirectory();

    /* Find operation mode */
    /*
          pcc-v2                          (a) directory chooser
          pcc-v2 -dir                     (b) directory chooser
          pcc-v2 -dir DIR                 (b) directory chooser
          pcc-v2 -dir     PLAYER          (b) directory chooser
          pcc-v2 -dir DIR PLAYER          (b) directory chooser

          pcc-v2      DIR                 (c) game if unique, otherwise chooser

          pcc-v2          PLAYER          (d) game
          pcc-v2      DIR PLAYER          (d) game
    */
    if (mode == mUnset) {
        if (player != 0) {
            /* case d */
            mode = mPlay;
        } else {
            if (getNumNonoptions() == 0) {
                /* case a */
                mode = mChooseGame;
            } else {
                /* case c */
                mode = mDefault;
            }
        }
    }
}
