/**
  *  \file client/screens/ionscreen.cc
  *  \brief Ion Storm Information Screen
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/ionscreen.h"
#include "client/tiles/ioninfo.h"
#include "client/tiles/objheader.h"
#include "cpluslib/format.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/ion.h"
#include "game/objtypes.h"
#include "gfx/palette.h"
#include "int/if/userif.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/window.h"
#include "util/nls.h"
#include "client/widgets/objcontrol.h"
#include "ui/widgets/button.h"

namespace {
    class WIonStormInfoWindow : public UIWindow {
        ObjectHolder h;
        GObjectType& type;
     public:
        WIonStormInfoWindow(GObjectType& type);
    };
}

/** Find nearest ion storm.
    \param pt starting point
    \return ion storm Id, or 0 */
int
findNearestIonStorm(GPoint pt)
{
    int found = 0;
    long dist2 = 0;
    bool inside = false;
    GIonStormType& type = global_ion_type;
    for (int iid = type.getNextNoWrap(0, false); iid != 0; iid = type.getNextNoWrap(iid, false)) {
        const GIonStorm& storm = type.getObject(iid);
        long ndist2 = distanceSquared(pt, storm.getPos());
        bool ninside = ndist2 <= square(storm.getRadius());
        /* Pick this storm if it is the first one, it is closer, or the point
           is inside this one but not inside the previous one */
        if (found <= 0 || ndist2 < dist2 || ninside > inside) {
            found = iid;
            dist2 = ndist2;
            inside = ninside;
        }
    }
    return found;
}

/************************** WIonStormInfoWindow **************************/

WIonStormInfoWindow::WIonStormInfoWindow(GObjectType& type)
    : UIWindow(_("Ion Storm Information"), BLUE_WINDOW, COLOR_GRAY),
      type(type)
{
    // VBox
    //   WObjectTypeHeaderTile
    //   WIonStormInfoTile
    //   HBox
    //     Close GoTo Forecast <--> H
    // Difference to PCC 1.x: Up/Down are in WObjectTypeHeaderTile
    setLayout(UIVBoxLayout::instance5);

    UIGroup& g1 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(h.add(new UIButton(0, _("Close"), SDLK_ESCAPE)).addStop());
    g1.add(h.add(new UIButton(1, _("Go To"), SDLK_RETURN)).addStop());
    g1.add(h.add(new UISpacer()));

    add(h.add(new WObjectTypeHeaderTile(type)));
    add(h.add(new WIonStormInfoTile(type)));
    add(g1);
    add(h.add(new WObjectTypeControl(type)));

    pack();
}

/** Ion storm information screen. Displays the ion storm currently
    selected in global_ion_type. */
void
doIonStormScreen()
{
    if (global_ion_type.getCount() == 0)
        return;

    WIonStormInfoWindow win(global_ion_type);
    ui_root->centerWidget(win);
    int result = ui_root->runModal(win);
    if (result == 1) {
        // "Go to"
        IntUserInterfaceBinding::getCurrentInstance()->doGoToChart(global_ion_type.getIonStorm().getPos());
    }
}
