/**
  *  \file client/tiles/shipequipment.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/shipequipment.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/ship.h"
#include "game/spec.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "game/config.h"
#include "game/cmds.h"
#include "game/global.h"
#include "ui/util.h"
#include "client/remote.h"
#include "game/host.h"
#include "client/wcolorframe.h"
#include "ui/widgets/button.h"

/*************************** WShipEquipmentTile **************************/

WShipEquipmentTile::WShipEquipmentTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, 6*font_heights[FONT_NORMAL]+3), type),
      cf_r(0),
      cf_g(0),
      cf_s(0)
{
    setLayout(RightJustified, 0, false);

    /* Some buttons: */
    int x = 0;
    if (host.isPHost() && config.CPEnableRemote()) {
        /* Remote control */
        cf_r = &h.add(new WColorFrame());
        cf_r->add(h.add(new UIButton("R", 'r')));
        addControl(GfxRect(x, 0, 24, 24), *cf_r);
        x += 25;
    }
    if (host.isPHost() && config.CPEnableGive()) {
        /* Give ship */
        cf_g = &h.add(new WColorFrame());
        cf_g->add(h.add(new UIButton("G", 'g')));
        addControl(GfxRect(x, 0, 24, 24), *cf_g);
        x += 25;
    }

    /* Ship spec sheet */
    cf_s = &h.add(new WColorFrame());
    cf_s->add(h.add(new UIButton("S", 's')));
    addControl(GfxRect(x, 0, 24, 24), *cf_s);
}

WShipEquipmentTile::~WShipEquipmentTile()
{ }

void
WShipEquipmentTile::drawShipEquipmentTile(UIAbstractTableWriter& tw)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Engines:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Primary:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Secondary:"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    tw.writeFixedCell(UIColor::tc_Static, _("Damage:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Crew:"));

    tw.newLColumn();

    const GShip* s = dynamic_cast<GShip*>(getObject());
    if (!s || !s->getHullId())
        return;

    const GShip& ship = *s;
    const GHull& hull = ship.getHull();

    /* engines */
    if (ship.getEngine().isKnown()) {
        string_t en = getEngine(ship.getEngine()).getName();
        if (ship.canDoSpecial(hf_Gravitonic))
            en = format(_("Grav. %s"), en);
        if (hull.getNumEngines() > 1)
            en = format("%d x %s", hull.getNumEngines(), en);
        tw.writeVariableCell(UIColor::tc_Green, en);
    } else {
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    /* primary */
    if (ship.getBeamCount().isKnown() && ship.getBeamType().isKnown()) {
        if (ship.getBeamCount() && ship.getBeamType()) {
            string_t bm = getBeam(ship.getBeamType()).getName();
            if (ship.getBeamCount() != 1)
                bm = format("%d x %s", ship.getBeamCount(), bm);
            tw.writeVariableCell(UIColor::tc_Green, bm);
        } else {
            tw.writeVariableCell(UIColor::tc_Green, _("none"));
        }
    } else {
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    /* secondary */
    if (!(ship.getTorpType().isKnown() && ship.getTorpLauncherCount().isKnown() && ship.getBayCount().isKnown())) {
        tw.writeVariableCell(UIColor::tc_Green, string_t());
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    } else if (ship.getTorpType() && ship.getTorpLauncherCount()) {
        string_t tl = format(_("%d x %s launcher%!1{s%}"),
                             ship.getTorpLauncherCount(),
                             getTorp(ship.getTorpType()).getName());
        string_t tc = (ship.getAmmo()
                       ? string_t(format(_("%d torpedo%!1{es%} aboard"), ship.getAmmo()))
                       : string_t(_("no torpedoes aboard")));
        tw.writeVariableCell(UIColor::tc_Green, tl);
        tw.writeVariableCell(UIColor::tc_Green, tc);
    } else if (ship.getBayCount()) {
        string_t tl = format(_("%d fighter bay%!1{s%}"), ship.getBayCount());
        string_t tc = (ship.getAmmo()
                       ? string_t(format(_("%d fighter%!1{s%} ready"), ship.getAmmo()))
                       : string_t(_("no fighters aboard")));
        tw.writeVariableCell(UIColor::tc_Green, tl);
        tw.writeVariableCell(UIColor::tc_Green, tc);
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    /* damage */
    if (!ship.getDamage().isKnown()) {
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    } else if (ship.getDamage()) {
        int i = ship.getSupplies() / 5;
        int j = ship.getDamage() - i;
        UISkin::Color c = UIColor::tc_Red;
        if (j <= 0) {
            j = 0;
            c = UIColor::tc_Yellow;
        }
        if (i)
            tw.writeVariableCell(c, format("%d%% (-%d: %d%%)", ship.getDamage(), i, j));
        else
            tw.writeVariableCell(c, format("%d%%", ship.getDamage()));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
    }

    /* crew */
    if (!ship.getCrew().isKnown()) {
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    } else if (ship.getCrew() != hull.getCrew()) {
        tw.writeVariableCell(ship.getCrew() > hull.getCrew() ? UIColor::tc_Yellow : UIColor::tc_Red,
                             format(_("%d of %d"), ship.getCrew(), hull.getCrew()));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, numToString(ship.getCrew()));
    }
    tw.finish();
}

void
WShipEquipmentTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(5, 0), FONT_NORMAL);
    drawShipEquipmentTile(tw);

    /* Some frames: */
    GShip* sh = dynamic_cast<GShip*>(getObject());
    if (sh != 0) {
        GCommandContainer& con = getDisplayedTurn().getCommands(sh->getOwner());
        if (cf_g != 0) {
            /* "give" button */
            cf_g->setColor((con.getCommand(GCommand::phc_GiveShip, sh->getId()) != 0
                            ? cf_Yellow
                            : cf_None));
        }
        if (cf_r != 0) {
            /* "remote" button */
            cf_r->setColor(getRemoteControlColor(getNewRemoteControlState(*sh, con)));
        }
        if (cf_s != 0) {
            /* "spec" button */
            cf_s->setColor(sh->hasAnyShipSpecificSpecials() ? cf_Green : cf_None);
        }
    } else {
        /* No ship selected */
        if (cf_g != 0)
            cf_g->setColor(cf_None);
        if (cf_r != 0)
            cf_r->setColor(cf_None);
        if (cf_s != 0)
            cf_s->setColor(cf_None);
    }
}

/************************ WNarrowShipEquipmentTile ***********************/

WNarrowShipEquipmentTile::WNarrowShipEquipmentTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(250, 5 * font_heights[FONT_NORMAL]), type)
{ }

void
WNarrowShipEquipmentTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();

    // Fetch the ship
    GShip* sh = dynamic_cast<GShip*>(getObject());
    if (!sh) {
        drawBackground(ctx, getInnerSize(0, 0));
        return;
    }

    // Coordinates
    int x = 0;
    int y = 0;
    int w = getInnerSize(0, 0).w;
    int dy = font_heights[FONT_NORMAL];

    // Engines
    if (sh->getEngine().isKnown()) {
        string_t en = getEngine(sh->getEngine()).getName();
        if (sh->canDoSpecial(hf_Gravitonic))
            en = format(_("Grav. %s"), en);
        if (sh->getHullId() > 0)
            if (sh->getHull().getNumEngines() > 1)
                en = format("%d x %s", sh->getHull().getNumEngines(), en);
        ctx.setColor(UIColor::tc_Green);
        outTextF(ctx, x, y, w, en);
    } else if (sh->getHullId() > 0) {
        string_t en;
        int n = sh->getHull().getNumEngines();
        if (sh->canDoSpecial(hf_Gravitonic))
            en = format(_("%d grav. engine%!1{s%}"), n);
        else
            en = format(_("%d engine%!1{s%}"), n);
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, en);
    } else {
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, string_t());
    }
    y += dy;

    // Beams
    if (sh->getBeamType().isKnown()) {
        if (sh->getBeamType() == 0 || (sh->getBeamCount().isKnown() && sh->getBeamCount() == 0)) {
            ctx.setColor(UIColor::tc_Green);
            outTextF(ctx, x, y, w, _("no beam weapons"));
        } else if (sh->getBeamCount().isKnown()) {
            // type and count
            ctx.setColor(UIColor::tc_Green);
            outTextF(ctx, x, y, w, format(_("%!d%!1{%0$d x %}%1$s"), sh->getBeamCount(), getBeam(sh->getBeamType()).getName()));
        } else if (sh->getHullId() > 0) {
            // just type
            ctx.setColor(UIColor::tc_Static);
            outTextF(ctx, x, y, w, format(_("up to %d %s"), sh->getHull().getMaxBeams(), getBeam(sh->getBeamType()).getName()));
        } else {
            ctx.setColor(UIColor::tc_Static);
            outTextF(ctx, x, y, w, getBeam(sh->getBeamType()).getName());
        }
    } else {
        if (sh->getHullId() > 0) {
            ctx.setColor(UIColor::tc_Static);
            if (sh->getHull().getMaxBeams() == 0) {
                outTextF(ctx, x, y, w, _("no beam weapons"));
            } else {
                outTextF(ctx, x, y, w, format(_("up to %d beam%!1{s%}"), sh->getHull().getMaxBeams()));
            }
        } else {
            ctx.setColor(UIColor::tc_Static);
            outTextF(ctx, x, y, w, string_t());
        }
    }
    y += dy;

    // Torps/Fighters
    if (sh->getTorpLauncherCount().isKnown() && sh->getTorpLauncherCount() > 0) {
        // Ship has torpedoes
        if (sh->getTorpType().isKnown()) {
            ctx.setColor(UIColor::tc_Green);
            if (sh->getAmmo() >= 0) {
                outTextF(ctx, x, y, w, format(_("%!d%!1{%0$d x %}%1$s, %d torpedo%!1{es%}"),
                                              sh->getTorpLauncherCount(),
                                              getTorp(sh->getTorpType()).getName(),
                                              sh->getAmmo()));
            } else {
                outTextF(ctx, x, y, w, format(_("%!d%!1{%0$d x %}%1$s"),
                                              sh->getTorpLauncherCount(),
                                              getTorp(sh->getTorpType()).getName()));
            }
        } else {
            ctx.setColor(UIColor::tc_Static);
            if (sh->getAmmo() >= 0) {
                outTextF(ctx, x, y, w, format(_("%d launcher%!1{s%}, %d torpedo%!1{es%}"),
                                              sh->getTorpLauncherCount(),
                                              sh->getAmmo()));
            } else {
                outTextF(ctx, x, y, w, format(_("%d torpedo launcher%!1{s%}"),
                                              sh->getTorpLauncherCount()));
            }
        }
    } else if (sh->getBayCount().isKnown() && sh->getBayCount() > 0) {
        // Ship has fighters
        ctx.setColor(UIColor::tc_Green);
        if (sh->getAmmo() >= 0) {
            outTextF(ctx, x, y, w, format(_("%d bay%!1{s%}, %d fighter%!1{s%}"),
                                          sh->getBayCount(),
                                          sh->getAmmo()));
        } else {
            outTextF(ctx, x, y, w, format(_("%d fighter bay%!1{s%}"),
                                          sh->getBayCount()));
        }
    } else if (sh->getTorpLauncherCount().isKnown() && sh->getBayCount().isKnown()) {
        // No torpedo launchers or fighters
        ctx.setColor(UIColor::tc_Green);
        outTextF(ctx, x, y, w, _("no secondary weapons"));
    } else if (sh->getHullId() > 0) {
        // Hull is known
        ctx.setColor(UIColor::tc_Static);
        if (sh->getHull().getMaxLaunchers() > 0) {
            outTextF(ctx, x, y, w, format(_("up to %d torpedo launcher%!1{s%}"), sh->getHull().getMaxLaunchers()));
        } else if (sh->getHull().getNumBays() > 0) {
            outTextF(ctx, x, y, w, format(_("%d fighter bay%!1{s%}"),
                                          sh->getHull().getNumBays()));
        } else {
            outTextF(ctx, x, y, w, _("no secondary weapons"));
        }
    } else {
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, string_t());
    }
    y += dy;

    // Crew [of Crew]
    if (sh->getCrew().isKnown()) {
        if (sh->getHullId() > 0 && sh->getCrew() != sh->getHull().getCrew()) {
            ctx.setColor(UIColor::tc_Red);
            outTextF(ctx, x, y, w, format(_("%d crew (of %d)"), sh->getCrew(), sh->getHull().getCrew()));
        } else {
            ctx.setColor(UIColor::tc_Green);
            outTextF(ctx, x, y, w, format(_("%d crew"), sh->getCrew()));
        }
    } else if (sh->getHullId() > 0) {
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, format(_("up to %d crew"), sh->getHull().getCrew()));
    } else {
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, string_t());
    }
    y += dy;

    // Damage
    if (sh->getDamage().isKnown()) {
        if (sh->getDamage() == 0)
            ctx.setColor(UIColor::tc_Green);
        else
            ctx.setColor(UIColor::tc_Red);
        outTextF(ctx, x, y, w, format(_("%d%% damage"), sh->getDamage()));
    } else {
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, string_t());
    }
    y += dy;
}
