/**
  *  \file game/cost.h
  *  \brief Cost/Resource Set
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_COST_H
#define PCC_V2_GAME_COST_H

#include <iosfwd>
#include "game/struct.h"
#include "game/cargotype.h"
#include "cpluslib/string.h"

/** Resource amount/cost. Represents the resource cost of a unit
    (weapon, building), or a corresponding resource set, and provides
    operations for manipulating it.

    This amount can be <em>valid</em> or <em>invalid</em>. It becomes
    invalid if a member is negative, that is, if a large value (the
    price) is subtracted from a small one (available resource). */
class GCost {
    int32 t, d, m, mc, sup;

    void trySupplySale();
 public:
    GCost();
    GCost(const TCost& cost);
    GCost& operator+=(const GCost& other);
    GCost& operator-=(const GCost& other);
    GCost& operator*=(int n);
    GCost  operator*(int n) const;
    bool   operator==(const GCost& other) const;
    bool   operator!=(const GCost& other) const
        { return !operator==(other); }
    int32 limitAmount(int32 amount, const GCost& item) const;

    void set(GCargoType type, int32 n);
    int32 get(GCargoType type) const;
    void add(GCargoType type, int32 n);
    void remove(GCargoType type, int32 n);

    bool isEnoughFor(const GCost& other) const;
    bool isValid() const;
    bool isEmpty() const;

    string_t toPHostString() const;
    string_t toCargospec() const;

    static bool isValidType(GCargoType t);

    static GCost fromString(string_t value);
};

/** Check whether cargo type is valid index to GCost, i.e.\ whether
    a GCost object can contain stuff of type \c t. */
inline bool
GCost::isValidType(GCargoType t)
{
    return (t == el_Tritanium || t == el_Duranium || t == el_Molybdenum
            || t == el_Money || t == el_Supplies);
    
}

std::ostream& operator<<(std::ostream&, const GCost&);

#endif
