/**
  *  \file game/planet-form.cc
  *  \brief Planet-related Formulas
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "game/planet.h"
#include "util/nls.h"
#include "cpluslib/misc.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/host.h"
#include "game/planet-int.h"

static inline double
pdiv(int a, int b)
{
    if (host.isPHost())
        return double(a) / b;
    else
        return a / b;
}

/** Maximum number of structures.
    \param clans      clans on planet
    \param threshold  below threshold, growth is linear; above it's rootic :-) */
int
GPlanet::getMaxStructuresFormula(int32 clans, int threshold)
{
    if (clans <= threshold)
        return clans;
    else
        /* the rounding function is actually ERnd, but that never
           makes a difference; sqrt(INT) never ends in .5 exactly */
        return threshold + int(std::sqrt(double(clans - threshold)) + 0.5);
}

/** Maximum number of structures.
    \param ps     Structure to query
    \param clans  Number of clans to assume
    \return Maximum allowed structures of that type */
int
GPlanet::getMaxStructures(TPlanetaryStructure ps, int32 clans) const
{
    switch (ps) {
     case str_Factories:
        return getMaxStructuresFormula(clans, 100);
     case str_Mines:
        return getMaxStructuresFormula(clans, 200);
     case str_Defense:
        return getMaxStructuresFormula(clans, 50);
     case str_BaseDefense:
        return base_info ? config.MaximumDefenseOnBase(base_info->data.owner) : 0;
    }
    return 0;
}

/** Colonists change.
    \param tax   desired tax rate
    \param mifa  mines plus factories on planet */
mn32_t
GPlanet::getColonistChange(int tax, int mifa) const
{
    /* can we compute that? */
    if (!getColonists().isKnown() || !isOwnerKnown() || !getTemperature().isKnown())
        return mn32_t();

    double common = 1000 - 80*tax - std::sqrt(double(getColonists()));
    bool   crystal = (raceId(getOwner()) == 7) && config.CrystalsPreferDeserts();
    if (host.isPHost()) {
        if (crystal)
            return mn32_t(int32(common - mifa/3.0 - (100-getTemperature())/0.66) / 100);
        else
            return mn32_t(int32(common - mifa/3.0 - std::abs(getTemperature()-50)/0.33) / 100);
    } else {
        if (crystal)
            return mn32_t(int32(common - (mifa/3 + 3*(100-getTemperature()))) / 100);
        else
            return mn32_t(int32(common - (mifa/3 + 3*std::abs(getTemperature()-50))) / 100);
    }
}

/** Colonists happiness change on this planet.
    \returns change on this planet in current situation */
mn32_t
GPlanet::getColonistChange() const
{
    if (!getColonistTax().isKnown() || !getStructures(str_Mines).isKnown() || !getStructures(str_Factories).isKnown())
        return mn32_t();
    else
        return getColonistChange(getColonistTax(), getStructures(str_Mines) + getStructures(str_Factories));
}

/** Get colonist tax amount, as requested. This returns the amount we want
    from the colonists, not the amount we'll finally get.
    \pre isOwnerKnown()
    \pre getColonists().isKnown() */
int32
GPlanet::getColonistDue(const int tax) const
{
    // Note that these formulas differ in rounding only. PHost uses
    // `Round', THost uses `ERnd' aka `I-don't-care-how-it-rounds'.
    ASSERT(isOwnerKnown() && getColonists().isKnown());
    const int rate = config.ColonistTaxRate(getOwner());
    if (host.isPHost()) {
        return radiv(radiv(getColonists() * tax, 1000) * rate, 100);
    } else {
        return rdivaddl(rdivaddl(getColonists() * tax, 1000, 0) * rate, 100, 0);
    }
}

/** Get colonist tax income.
    \param tax     [in] tax rate
    \param rem_inc [out] this is set to the amount of taxes which we can
    still get from the natives (regarding MaxPlanetaryIncome)
    \returns megacredits collected.
    \pre isOwnerKnown()
    \pre getColonists().isKnown() */
int32
GPlanet::getColonistDueLimited(const int tax, int32& rem_inc) const
{
    const int32 due = getColonistDue(tax), max = config.MaxPlanetaryIncome(getOwner());
    if (due < max) {
        rem_inc = max - due;
        return due;
    } else {
        rem_inc = 0;
        return max;
    }
}

/** Maximum population on planet, for a race.
    \param race player number
    \returns maximum population in clans. */
mp32_t
GPlanet::getMaxSupportedColonists(int player) const
{
    bool crystal = (raceId(player) == 7 && config.CrystalsPreferDeserts());
    int32 limit;
    if (host.isPHost()) {
        if (!config.ClimateLimitsPopulation())
            return mp32_t(250000);
        if (!getTemperature().isKnown())
            return mp32_t();
        if (crystal) {
            if (config.CrystalSinTempBehavior())
                if (getTemperature() >= 15)
                    return mp32_t(int32(100000 * std::sin(getTemperature() * PI / 100)));
                else
                    return mp32_t(3 + getTemperature() * config.MaxColTempSlope() / 100);
            else
                return mp32_t(1000 * getTemperature());
        }

        if (getTemperature() >= 85)
            limit = config.MaxColTempSlope() * (100-getTemperature()) / 100 + 1;
        else if (getTemperature() <= 14)
            limit = config.MaxColTempSlope() * getTemperature() / 100 + 3;
        else
            limit = int32(100000 * std::sin(getTemperature() * PI / 100));
    } else {
        if (!config.ClimateLimitsPopulation())
            return mp32_t(100000);
        if (!getTemperature().isKnown())
            return mp32_t();
        if (crystal)
            return mp32_t(1000 * getTemperature());

        if (getTemperature() >= 85)
            limit = 2 * (100 - getTemperature()) + 1;
        else if (getTemperature() <= 14)
            limit = 2 * getTemperature() + 3;
        else
            limit = int32(100000 * std::sin((100 - getTemperature()) * 0.0314) + 0.5);
    }

    if (raceId(player) == 10 && getTemperature() <= 19 && limit < 90000)
        limit = 90000;
    if ((raceId(player) == 4 || raceId(player) >= 9) && getTemperature() >= 84 && limit < 60)
        limit = 60;

    return mp32_t(limit);
}

/* Natives */

/** Native happiness change.
    \param tax   desired tax rate
    \param mifa  mines plus factories on planet */
mn32_t
GPlanet::getNativeChange(int tax, int mifa) const
{
    /* formula is correct for PHost and THost */
    if (!getNativeGovernment().isKnown() || !getNatives().isKnown())
        return mn32_t();

    int change =
        int32(500 + 50 * getNativeGovernment()
              - 85 * tax
              - pdiv(mifa, 2)
              - std::sqrt(double(getNatives())))
        / 100;
    if (getNativeRace().isKnown() && getNativeRace() == nr_Avian)
        change += 10;
    return mn32_t(change);
}

/** Native happiness change.
    \returns happiness change for current situation. */
mn32_t
GPlanet::getNativeChange() const
{
    if (!getNativeTax().isKnown() || !getStructures(str_Mines).isKnown() || !getStructures(str_Factories).isKnown())
        return mn32_t();
    else
        return getNativeChange(getNativeTax(), getStructures(str_Mines) + getStructures(str_Factories));
}

/** Get native tax amount, as requested, for this planet. This returns
    the amount we're asking from them, not what we'll get.
    \pre native race, government, population, and planet owner known */
int32
GPlanet::getNativeDue(int tax) const
{
    ASSERT(getNativeRace().isKnown() && getNativeGovernment().isKnown());
    ASSERT(getNatives().isKnown() && isOwnerKnown());

    return getNativeDue(tax, getNativeRace(), getNativeGovernment(),
                        getNatives(), getOwner());
}

/** Get native tax amount, as requested, parameterized. This returns
    the amount we're asking from them, not what we'll get.
    \param tax    tax rate
    \param race   native race
    \param gov    government factor (SPI)
    \param pop    native clans
    \param owner  owner (for production rates) */
int32
GPlanet::getNativeDue(int tax, int race, int gov, int32 pop, int owner)
{
    int32 due;
    if (host.isPHost()) {
        due = radiv(radiv(tax * gov * pop, 5000) * config.NativeTaxRate(owner), 100);
    } else {
        due = rdivaddl(rdivaddl(tax * gov * pop, 5000, 0) * config.ColonistTaxRate(owner), 100, 0);
    }
    if (race == nr_Insectoid)
        return 2*due;
    else
        return due;
}

/** Get native tax amount, limited. Limits the tax income by income
    limit and available colonists.
    \param tax     tax rate
    \param rem_inc remaining allowed income; output of getColonistDueLimited(). */
int32
GPlanet::getNativeDueLimited(int tax, int32 rem_inc) const
{
    /* amorphs don't pay */
    if (getNativeRace() == nr_Amorphous)
        return 0;

    /* cyborgs can only tax to 20 */
    int limit = host.getNativeTaxRateLimit(getOwner());
    if (tax > limit)
        tax = limit;

    /* normal formulas here */
    int32 due = getNativeDue(tax);
    int32 colos = getColonists();

    if (host.isPHost()) {
        if (getNativeRace() == nr_Insectoid)
            colos *= 2;
        colos = radiv(colos * config.NativeTaxRate(getOwner()), 100);
    } else {
        colos = colos * config.ColonistTaxRate(getOwner()) / 100;
        if (getNativeRace() == nr_Insectoid)
            colos *= 2;
    }

    if (due > colos)
        due = colos;
    if (due > rem_inc)
        due = rem_inc;
    return due;
}

/** Get Bovinoid supply contribution. Contribution is not yet limited by
    colonists. */
int32
GPlanet::getBovinoidSupplyContribution() const
{
    // FIXME: we now return 0 if the value is unknown
    if (getNativeRace().isKnown() && isOwnerKnown() && getNatives().isKnown())
        if (getNativeRace() == nr_Bovinoid)
            return getBovinoidSupplyContribution(getNatives(), getOwner());
        else
            return 0;
    else
        return 0;
}

/** Get Bovinoid supply contribution. This returns the amount of
    supplies finally collected. */
int32
GPlanet::getBovinoidSupplyContributionLimited() const
{
    // FIXME: the same problem as in getBovinoidSupplyContribution(int,int)
    // appears here. We *could* add a workaround here (compute totals,
    // compute factory contribution, compute difference), but is it
    // really worth to do that for an accuracy of +/-1 with a config
    // option no-one uses?
    //      -- update: it's going to be used in PList 2.5, so someday
    //         it should be supported
    if (!getNativeRace().isKnown() || !getColonists().isKnown())
        return 0;

    if (getNativeRace() == nr_Bovinoid) {
        int32 limit = host.isPHost() ? getColonists() * config.ProductionRate(getOwner()) / 100 : getColonists();
        int32 due = getBovinoidSupplyContribution();
        if (due < limit)
            return due;
        else
            return limit;
    } else {
        return 0;
    }
}

/** Get Amorphous colonist breakfast. Returns the number of clans
    eaten by our lovely amorphous natives when they have a happiness
    as specified. */
int32
GPlanet::getAmorphousBreakfast(int happy)
{
    if (host.isPHost()) {
        if (happy >= 70)
            return 5;
        else if (happy >= 50)
            return 20;
        else
            return 40;
    } else {
        if (happy > 90)
            return 5;
        else
            return 95 - happy;
    }
}

/** Get Bovinoid supply contribution, parameterized. */
int32
GPlanet::getBovinoidSupplyContribution(int32 pop, int owner)
{
    // FIXME: for ultimate accuracy, we have to consider factories as
    // well. PHost does "(bovi + factories) * ProductionRate", not
    // "bovi * ProductionRate + factories * ProductionRate".
    if (host.isPHost())
        return (pop / 100) * config.ProductionRate(owner) / 100;
    else
        return pop / 100;
}

/** Get mining capacity.
    \param index Mineral index
    \param mines Number of mines on planet
    \pre density is known */
int
GPlanet::getMiningCapacity(int index, int mines) const
{
    // Gather inputs
    int density = getOreDensity(index);

    int mining_rate;
    if (isOwnerKnown()) {
        mining_rate = config.RaceMiningRate(getOwner());
    } else {
        mining_rate = 100;
    }

    int reptile_factor;
    if (getNativeRace().isKnown()
        && getNativeRace() == nr_Reptilian
        && (!getNatives().isKnown() || getNatives() > 0))
    {
        reptile_factor = 2;
    } else {
        reptile_factor = 1;
    }

    // Host-dependant formula
    if (!host.isPHost()) {
        return rdivaddl(rdivaddl(density * int32(mines), 100, 0) * mining_rate, 100, 0) * reptile_factor;
    } else {
        int add = (host.isPHostRoundingMiningResults() ? 50 : 0);
        return (((density * int32(mining_rate) + add) / 100) * reptile_factor * mines + add) / 100;
    }
}
