/**
  *  \file game/prefio.cc
  *  \brief User Preferences I/O
  *
  *  FIXME: having this in a separate module means that checkinits.pl
  *  will not warn of a missing mod_pref.
  *
  *  Up to beta 12, getUserSettingsDirectory() was implemented to
  *  return $HOME resp. %APPDATA%. Since beta 13, it is a subdirectory
  *  of said folder (.pcc2 under Unix, PCC2 under Windows). Right now,
  *  we load the file from both locations, and save it to where we got
  *  it from. This means we keep interchangability with previous
  *  versions, in case a user doesn't like this version and goes back
  *  to the previous one.
  *
  *  \todo A future version should migrate the pcc2.ini file into the
  *  subdirectory.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/pref.h"
#include "arch/platform.h"
#include "cpluslib/format.h"
#include "io/dirs.h"
#include "io/memstream.h"
#include "io/streams.h"
#include "util/confpars.h"
#include "util/console.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "util/except.h"

static int loaded_pref_version;

static string_t
getPrefFileName(int version)
{
    return Directory::makeFileName(Platform::getUserSettingsDirectory(version), "pcc2.ini");
}

static void
createPath(string_t name)
{
    string_t parent = Directory::getDirName(name);
    string_t child  = Directory::getFileName(name);

    try {
        Ptr<Directory> d = Directory::openDir(parent);
        Ptr<DirectoryEntry> e = d->getEntryByName(child);
        if (e->getFileType() != DirectoryEntry::FT_DIRECTORY)
            e->createAsDirectory();
    }
    catch (FileProblemException& e) {
        console.write(LOG_ERROR, format("%s: %s", e.getFileName(), e.what()));
    }
}

/** Create directory name for preference files. */
void
createUserPreferencesDirectory()
{
    createPath(Platform::getUserSettingsDirectory(1));
}

/** Load user preferences file (pcc2.ini). */
static void
loadUserPreferences()
{
    int version = 1;
    Ptr<Stream> s = Stream::openFileNT(getPrefFileName(version), Stream::C_READ);
    if (!s) {
        version = 0;
        s = Stream::openFileNT(getPrefFileName(version), Stream::C_READ);
    }
    if (s) {
        ConfigFileParser cfp(getUserPreferences(), true, ConfigOption::srcUser);
        cfp.setUtf8();
        cfp.parseFile(*s);
        loaded_pref_version = version;
    } else {
        loaded_pref_version = 1;
    }
}

/** Save user preferences file (pcc2.ini). */
static void
saveUserPreferences()
{
    string_t name = getPrefFileName(loaded_pref_version);

    // Update config file in memory
    MemoryStream ms;
    ConfigFileUpdater cfu(getUserPreferences(), ms);
    if (Ptr<Stream> s = Stream::openFileNT(name, Stream::C_READ)) {
        cfu.parseFile(*s);
    }
    cfu.finish();

    // Create path
    if (loaded_pref_version > 0) {
        createUserPreferencesDirectory();
    }

    // Write it out
    Ptr<Stream> s = Stream::openFileNT(name, Stream::C_CREATE | Stream::C_WRITE);
    if (!s) {
        console.write(LOG_ERROR, format(_("Unable to create configuration file \"%s\""), name));
    } else {
        s->writeT(ms.getBuffer().ptr(), ms.getBuffer().getSize());
    }
}

extern const ModuleInitializer mod_pref[] = {
    { OnEnterNoGame | OnEnterDirectory, loadUserPreferences },
    { OnLeftDirectory,                  saveUserPreferences },
    { 0, 0 },
};
