/**
  *  \file gfx/anim.cc
  *  \brief Animation Implementation
  *
  *  Data structures and classes for animations.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "gfx/anim.h"

/** \class GfxAnimation
    \brief Animation

    This class manages an animation as a series of frames (pixmaps). */

/** Create Empty Animation. This can not be used until you add frames
    (pixmaps) to it. */
GfxAnimation::GfxAnimation()
    : wi(0), he(0)
{ }

GfxAnimation::~GfxAnimation()
{ }

/** Add one frame. Note that all pixmaps should have the same size.
    \param apix   pixmap to display
    \param anext  next pixmap to display (index), -1 for end of animation
    \param atime  how long to display this frame */
void
GfxAnimation::addFrame(Ptr<GfxPixmap> apix, int anext, int atime)
{
    data.push_back(GfxAniFrame(apix, anext, atime));
    if (wi == 0 && he == 0) {
        wi = apix->getWidth();
        he = apix->getHeight();
    }
}

/****************************** GfxAniSprite *****************************/

/** Constructor.
    \param x,y     where to display this animation
    \param alayer  layer of animation
    \param aanim   animation to display */
GfxAniSprite::GfxAniSprite(int x, int y, int z, Ptr<GfxAnimation> aanim)
    : GfxSprite3D(GfxRect(x, y, aanim->getWidth(), aanim->getHeight()), z),
      anim(aanim)
{
    setIndex(anim->getStart());
}

/** Move to animation frame /aip/. */
void
GfxAniSprite::setIndex(int aip)
{
    ip = aip;
    time = anim->getFrame(ip).time;
}

void
GfxAniSprite::tick()
{
    if (--time <= 0) {
        int new_ip = anim->getFrame(ip).next;
        if (new_ip < 0) {
            deleteThis();
        } else {
            setIndex(new_ip);
            update();
        }
    }
}

void
GfxAniSprite::draw(GfxCanvas& can)
{
    anim->getFrame(ip).pixmap->blit(can, getPos().x, getPos().y);
}
