/**
  *  \file int/expr.cc
  *  \brief Interpreter: Expression Parse Trees
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/expr.h"
#include "int/interror.h"
#include "int/unary.h"
#include "int/binary.h"

/****************************** IntExprNode ******************************/

void
IntExprNode::compileEffect(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    // Compute value and discard it
    compileValue(bco, cc);
    bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
}

void
IntExprNode::compileCondition(IntBytecodeObject& bco, const IntCompilationContext& cc, IntBytecodeObject::label_t ift, IntBytecodeObject::label_t iff)
{
    // Generate two-way jump
    compileValue(bco, cc);
    bco.addJump(IntOpcode::jIfTrue | IntOpcode::jPopAlways, ift);
    bco.addJump(IntOpcode::jAlways, iff);
}

/*************************** IntRValueExprNode ***************************/

void
IntRValueExprNode::compileStore(IntBytecodeObject& /*bco*/, const IntCompilationContext& /*cc*/, IntExprNode* /*rhs*/)
{
    throw IntError::notAssignable();
}

void
IntRValueExprNode::compileRead(IntBytecodeObject& /*bco*/, const IntCompilationContext& /*cc*/)
{
    throw IntError::notAssignable();
}

void
IntRValueExprNode::compileWrite(IntBytecodeObject& /*bco*/, const IntCompilationContext& /*cc*/)
{
    throw IntError::notAssignable();
}

/*************************** IntLiteralExprNode **************************/

IntLiteralExprNode::IntLiteralExprNode()
    : value(0)
{ }

IntLiteralExprNode::~IntLiteralExprNode()
{
    delete value;
}

/** Set value. IntLiteralExprNode will take ownership. */
void
IntLiteralExprNode::setValue(IntValue* value) throw()
{
    this->value = value;
}

void
IntLiteralExprNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& /*cc*/)
{
    bco.addPushLiteral(value);
}

/************************ IntSimpleRValueExprNode ************************/

IntSimpleRValueExprNode::IntSimpleRValueExprNode()
    : a(0),
      b(0),
      c(0)
{ }

IntSimpleRValueExprNode::~IntSimpleRValueExprNode()
{
    delete a;
    delete b;
    delete c;
}

/** Set child nodes for unary expression. IntSimpleRValueExprNode will take ownership. */
void
IntSimpleRValueExprNode::setUnary(IntExprNode* a) throw()
{
    this->a = a;
}

/** Set child nodes for binary expression. IntSimpleRValueExprNode will take ownership. */
void
IntSimpleRValueExprNode::setBinary(IntExprNode* a, IntExprNode* b) throw()
{
    this->a = a;
    this->b = b;
}

/** Set child nodes for ternary expression. IntSimpleRValueExprNode will take ownership. */
void
IntSimpleRValueExprNode::setTernary(IntExprNode* a, IntExprNode* b, IntExprNode* c) throw()
{
    this->a = a;
    this->b = b;
    this->c = c;
}

/*************************** IntSimpleExprNode ***************************/

IntSimpleExprNode::IntSimpleExprNode(IntOpcode::Major major, uint8_t minor)
    : major(major), minor(minor)
{ }

void
IntSimpleExprNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    if (a)
        a->compileValue(bco, cc);
    if (b)
        b->compileValue(bco, cc);
    if (c)
        c->compileValue(bco, cc);
    bco.addInstruction(major, minor, 0);
}

/**************************** IntCaseExprNode ****************************/

/** Constructor.
    \param minor Minor opcode for case-sensitive operation */
IntCaseExprNode::IntCaseExprNode(uint8_t minor)
    : minor(minor)
{ }

void
IntCaseExprNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    a->compileValue(bco, cc);
    b->compileValue(bco, cc);
    bco.addInstruction(IntOpcode::maBinary,
                       (cc.compilation_flags & IntCompilationContext::CaseBlind) ? minor+1 : minor,
                       0);
}

/** Convert to assignment. If this is an equality comparison, creates a new IntAssignmentNode
    containing our child nodes, and returns that. This is a destructive operation which will
    make this IntCaseExprNode unusable for compilation. You are expected to delete it and
    replace the pointer by the new element returned from this function.

    If this is not an equality comparison, does nothing but return 0. */
IntExprNode*
IntCaseExprNode::convertToAssignment()
{
    if (minor == biCompareEQ) {
        IntSimpleRValueExprNode* result = new IntAssignmentNode();
        result->setBinary(a, b);
        a = b = 0;
        return result;
    } else {
        return 0;
    }
}

/*************************** IntAssignmentNode ***************************/

void
IntAssignmentNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    a->compileStore(bco, cc, b);
}

/**************************** IntSequenceNode ****************************/

void
IntSequenceNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    a->compileEffect(bco, cc);
    b->compileValue(bco, cc);
}

/*************************** IntConditionalNode **************************/

void
IntConditionalNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    IntBytecodeObject::label_t ift = bco.makeLabel();
    IntBytecodeObject::label_t iff = bco.makeLabel();
    IntBytecodeObject::label_t end = bco.makeLabel();

    a->compileCondition(bco, cc, ift, iff);
    bco.addLabel(ift);
    b->compileValue(bco, cc);
    bco.addJump(IntOpcode::jAlways, end);
    bco.addLabel(iff);
    if (c != 0)
        c->compileValue(bco, cc);
    else
        bco.addPushLiteral(0);
    bco.addLabel(end);
}

/***************************** IntLogicalNode ****************************/

/** Constructor.
    \param shortcut_jump Condition for the shortcut jump
    \param binary_op     Minor opcode for the binary operation */
IntLogicalNode::IntLogicalNode(uint8_t shortcut_jump, uint8_t binary_op)
    : shortcut_jump(shortcut_jump),
      binary_op(binary_op)
{ }

void
IntLogicalNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    // We need the 'ubool' instruction to guarantee the result to be boolean even
    // when we take the shortcut. The only exception is Xor, where the shortcut
    // value is "singular".
    // a And b       a Or b        a Xor b
    // => a          a             a
    //    ubool      ubool        [ubool]
    //    jf fini    jt fini       je fini
    //    b          b             b
    //    band       bor           bxor
    // fini:      fini:          fini:

    IntBytecodeObject::label_t fini = bco.makeLabel();

    a->compileValue(bco, cc);
    if (binary_op != biXor)
        bco.addInstruction(IntOpcode::maUnary, unBool, 0);
    bco.addJump(shortcut_jump, fini);
    b->compileValue(bco, cc);
    bco.addInstruction(IntOpcode::maBinary, binary_op, 0);
    bco.addLabel(fini);
}

void
IntLogicalNode::compileCondition(IntBytecodeObject& bco, const IntCompilationContext& cc, IntBytecodeObject::label_t ift, IntBytecodeObject::label_t iff)
{
    // a And b       a Or b        a Xor b
    // => a          => a          => a
    //    jfep iff      jtp ift       je skip
    //    b             b             b
    //    jfep iff      jtp ift       bxor
    //    j ift         j iff      skip: jtp ift
    //                                j iff

    if (binary_op == biAnd) {
        // FIXME(?): this generates different side-effects from compileValue().
        // compileValue() will evaluate RHS if LHS is Empty (to distinguish Empty and False),
        // whereas this implementation will not. If we wish to keep efficient short-circuit
        // evaluation, it probably makes sense to keep this undefined.
        IntBytecodeObject::label_t x = bco.makeLabel();
        a->compileCondition(bco, cc, x, iff);
        bco.addLabel(x);
        b->compileCondition(bco, cc, ift, iff);
    } else if (binary_op == biOr) {
        IntBytecodeObject::label_t x = bco.makeLabel();
        a->compileCondition(bco, cc, ift, x);
        bco.addLabel(x);
        b->compileCondition(bco, cc, ift, iff);
    } else if (binary_op == biXor) {
        // This is harder than the others because we cannot use compileCondition here.
        // Jump threading also does not work, because we have no 'jump and pop only if
        // condition true' instruction. Since Xor is comparatively rare, I hope this
        // isn't too much of a problem.
        IntBytecodeObject::label_t x = bco.makeLabel();
        a->compileValue(bco, cc);
        bco.addJump(shortcut_jump, x);
        b->compileValue(bco, cc);
        bco.addInstruction(IntOpcode::maBinary, binary_op, 0);
        bco.addLabel(x);
        bco.addJump(IntOpcode::jIfTrue | IntOpcode::jPopAlways, ift);
        bco.addJump(IntOpcode::jAlways, iff);
    } else {
        IntSimpleRValueExprNode::compileCondition(bco, cc, ift, iff);
    }
}

/***************************** IntMemberNode *****************************/

/** Constructor.
    \param name Name of member */
IntMemberNode::IntMemberNode(string_t name)
    : name(name), expr(0)
{ }

IntMemberNode::~IntMemberNode()
{
    delete expr;
}

void
IntMemberNode::setExpression(IntExprNode* expr) throw()
{
    this->expr = expr;
}

void
IntMemberNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    // PUSHMEM nn      rr:R => rr:R.name[nn]
    expr->compileValue(bco, cc);
    bco.addInstruction(IntOpcode::maMemref, IntOpcode::miIMLoad, bco.addName(name));
}

void
IntMemberNode::compileStore(IntBytecodeObject& bco, const IntCompilationContext& cc, IntExprNode* rhs)
{
    // STOREMEM nn     rr:val:R => rr:val
    // 'val' already on stack, so we must compute 'R'
    rhs->compileValue(bco, cc);
    expr->compileValue(bco, cc);
    bco.addInstruction(IntOpcode::maMemref, IntOpcode::miIMStore, bco.addName(name));
}

void
IntMemberNode::compileRead(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    expr->compileValue(bco, cc);
    bco.addInstruction(IntOpcode::maStack,  IntOpcode::miStackDup, 0);
    bco.addInstruction(IntOpcode::maMemref, IntOpcode::miIMLoad, bco.addName(name));
}

void
IntMemberNode::compileWrite(IntBytecodeObject& bco, const IntCompilationContext& /*cc*/)
{
    bco.addInstruction(IntOpcode::maStack,  IntOpcode::miStackSwap, 1);
    bco.addInstruction(IntOpcode::maMemref, IntOpcode::miIMStore, bco.addName(name));
}

/*************************** IntIdentifierNode ***************************/

/** Constructor.
    \param name Name of identifier */
IntIdentifierNode::IntIdentifierNode(string_t name)
    : name(name)
{ }

IntIdentifierNode::~IntIdentifierNode()
{ }

void
IntIdentifierNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    bco.addVariableReferenceInstruction(IntOpcode::maPush, name, cc);
}

void
IntIdentifierNode::compileStore(IntBytecodeObject& bco, const IntCompilationContext& cc, IntExprNode* rhs)
{
    rhs->compileValue(bco, cc);
    bco.addVariableReferenceInstruction(IntOpcode::maStore, name, cc);
}

void
IntIdentifierNode::compileRead(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    bco.addVariableReferenceInstruction(IntOpcode::maPush, name, cc);
}

void
IntIdentifierNode::compileWrite(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    bco.addVariableReferenceInstruction(IntOpcode::maStore, name, cc);
}

/************************** IntFunctionCallNode **************************/

/** Append argument.
    \param arg Argument, newly-allocated */
void
IntFunctionCallNode::addArgument(IntExprNode* arg)
{
    args.push_back(arg);
}

/************************** IntIndirectCallNode **************************/

IntIndirectCallNode::IntIndirectCallNode()
    : func(0)
{ }

IntIndirectCallNode::~IntIndirectCallNode()
{
    delete func;
}

/** Set function. IntIndirectCallNode takes ownership. */
void
IntIndirectCallNode::setFunction(IntExprNode* func)
{
    this->func = func;
}

void
IntIndirectCallNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    // PUSHIND nargs    rr:args:R => rr:result
    for (ptr_vector<IntExprNode>::size_type i = 0; i != args.size(); ++i)
        args[i]->compileValue(bco, cc);
    func->compileValue(bco, cc);
    bco.addInstruction(IntOpcode::maIndirect, IntOpcode::miIMLoad + IntOpcode::miIMRefuseProcedures, args.size());
}

void
IntIndirectCallNode::compileStore(IntBytecodeObject& bco, const IntCompilationContext& cc, IntExprNode* rhs)
{
    // STOREIND nargs   rr:args:val:R => rr:val
    for (ptr_vector<IntExprNode>::size_type i = 0; i != args.size(); ++i)
        args[i]->compileValue(bco, cc);
    rhs->compileValue(bco, cc);
    func->compileValue(bco, cc);
    bco.addInstruction(IntOpcode::maIndirect, IntOpcode::miIMStore + IntOpcode::miIMRefuseProcedures, args.size());
}

void
IntIndirectCallNode::compileRead(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    // Compute inputs            => ...:args:func
    for (ptr_vector<IntExprNode>::size_type i = 0; i != args.size(); ++i)
        args[i]->compileValue(bco, cc);
    func->compileValue(bco, cc);

    // Duplicate everything      => ...:args:func:args:func
    uint32_t nwords = args.size()+1;
    for (uint32_t i = 0; i < nwords; ++i)
        bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDup, nwords-1);

    // Read                      => ...:args:func:value
    bco.addInstruction(IntOpcode::maIndirect, IntOpcode::miIMLoad + IntOpcode::miIMRefuseProcedures, args.size());
}

void
IntIndirectCallNode::compileWrite(IntBytecodeObject& bco, const IntCompilationContext& /*cc*/)
{
    // We have ...:args:func:value,
    // we need ...:args:value:func
    bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackSwap, 1);
    bco.addInstruction(IntOpcode::maIndirect, IntOpcode::miIMStore + IntOpcode::miIMRefuseProcedures, args.size());
}
