/**
  *  \file ui/prefix.cc
  *  \brief Prefix Argument Handling
  *
  *  How the prefix argument is handled:
  *  - it is associated with key events (UIEvent::key.arg)
  *  - it is associated with user commands (UICommand::arg)
  *  - mouse event handlers have to poll it itself
  *  - UIRoot::run cancels the prefix argument when it is not
  *    consumed after a mouse release.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/prefix.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "ui/window.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"

/** Popup for entering a prefix argument. Contains the user-interface
    logic for entering prefix arguments.

    Users can type a sequence of digits to produce a prefix argument
    of up to four decimals. In addition, we allow
    - backspace to correct
    - ESC to discard the prefix
    - '*' and '/' operators to allow simple mathematics */
class UIPrefixPopup : public UIFixedWidget {
    UIPrefixArgInput inp;

    bool blink;                 ///< Cursor blink state.
    int  blink_count;           ///< Cursor blink timer divider.

    void update();
 public:
    UIPrefixPopup(int value);
    ~UIPrefixPopup();

    bool handleEvent(const UIEvent& event, bool second_pass);
    void drawContent(GfxCanvas& can);
};

UIPrefixArgInput::UIPrefixArgInput(int value)
    : value(value),
      second_value(0),
      op(NoOp)
{ }

/** Get current text. Formats the input into the text shown to the user. */
string_t
UIPrefixArgInput::getText() const
{
    const char* title = _("Prefix: ");
    if (op != NoOp) {
        char opchar = (op == MultiplyOp ? '*' : '/');
        if (second_value)
            return format("%s%d%c%d", title, value, opchar, second_value);
        else
            return format("%s%d%c", title, value, opchar);
    }
    return format("%s%d", title, value);
}

/** Get current value of this prefix argument. */
int
UIPrefixArgInput::getValue() const
{
    if (second_value != 0) {
        if (op == MultiplyOp)
            return value * second_value;
        else if (op == DivideOp)
            return value / second_value;
        else
            return value;
    } else {
        return value;
    }
}

UIPrefixArgInput::Action
UIPrefixArgInput::handleKey(uint32_t key)
{
    uint32_t raw_key = key & ~ss_Alt;
    if (raw_key >= '0' && raw_key <= '9') {
        int* which_value = (op == NoOp ? &value : &second_value);
        int old_value = *which_value;
        /* we only allow entry of a number if
           a. the resulting component is less than 10000
           b. the resulting value is less than 10000
           c. the resulting value is not zero */
        if (old_value >= 1000)
            return Accepted;
        *which_value = 10 * old_value + (raw_key - '0');
        if (getValue() == 0 || getValue() >= 10000) {
            *which_value = old_value;
            return Accepted;
        }
        /* okay, we accepted the number */
        return Accepted;
    } else if (raw_key == '*' || raw_key == '/') {
        /* when entering '*',
           '77/11' goes to '7*'
           '77/'   goes to '77*' */
        value = getValue();
        second_value = 0;
        op = (raw_key == '*' ? MultiplyOp : DivideOp);
        return Accepted;
    } else if (raw_key == SDLK_BACKSPACE) {
        /* 77/11 goes to 77/1
           77/1  goes to 77/
           77/   goes to 77
           77    goes to 7 */
        if (op != NoOp) {
            if (second_value > 0)
                second_value /= 10;
            else
                op = NoOp;
        } else {
            value /= 10;
            if (!value) {
                return Canceled;
            }
        }
        return Accepted;
    } else if (raw_key == SDLK_ESCAPE) {
        return Canceled;
    } else {
        return NotHandled;
    }
}

/** Create popup.
    \param value Initial value */
UIPrefixPopup::UIPrefixPopup(int value)
    : inp(value),
      blink(false),
      blink_count(0)
{
    update();
}

UIPrefixPopup::~UIPrefixPopup()
{ }

/** Update position on screen. Computes the new position, moves the
    widget if needed, and requests redraw. */
void
UIPrefixPopup::update()
{
    string_t text = inp.getText();
    GfxRect new_extent(0, 0,
                       fonts[FONT_FIXED]->getTextWidth(text) + 14,
                       fonts[FONT_FIXED]->getTextHeight(text) + 2);
    new_extent.moveToEdge(ui_root->getExtent(), 1, 1, 0);
    setExtent(new_extent, false);
    drawWidget();
}

bool
UIPrefixPopup::handleEvent(const UIEvent& event, bool second_pass)
{
    switch (event.type) {
     case ev_Key: {
        switch (inp.handleKey(event.key.code)) {
         case UIPrefixArgInput::Accepted:
            update();
            return true;
         case UIPrefixArgInput::Canceled:
            stop(0);
            return true;
         default:
            /* possibly-defined key: note prefix arg, and exit */
            if (getClassOfKey(event.key.code & ~ss_AllMod) == KeyNormal) {
                UIEvent newEvent = event;
                newEvent.key.arg = inp.getValue();
                stop(0);
                ungetEvent(newEvent);
            }
            return true;
        }
     }

     case ev_Mouse:
        /* mouse events with a pressed button cause the prefix
           argument to be accepted; it will usually be associated with
           the next button release (see UIRoot::run()). */
        if (event.mouse.b != 0) {
            setPrefixArg(inp.getValue());
            stop(0);
            ungetEvent(event);
        }
        return true;

     case ev_Tick:
        if (++blink_count > 40) {
            blink = !blink;
            blink_count = 0;
            drawWidget(false);
        }
        /* FALLTHROUGH */

     default:
        /* handle Idle and Tick events to parent */
        return UIFixedWidget::handleEvent(event, second_pass);
    }
}

void
UIPrefixPopup::drawContent(GfxCanvas& can)
{
    const GfxRect& r = getExtent();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    // FIXME: define this color elsewhere? It's the same as used
    // in PCC 1.x, a rather bright yellow.
    drawSolidBar(ctx, r, COLOR_FIRE + 29);

    ctx.setColor(COLOR_DARKYELLOW);
    drawHLine(ctx, r.x, r.getBottomY()-1, r.getRightX()-1);
    drawVLine(ctx, r.getRightX()-1, r.y, r.getBottomY()-2);

    ctx.setColor(COLOR_FIRE + 30);
    drawHLine(ctx, r.x+1, r.y, r.getRightX()-1);
    drawVLine(ctx, r.x, r.y, r.getBottomY()-2);

    ctx.setColor(COLOR_BLACK);
    ctx.useFont(*fonts[FONT_FIXED]);
    outText(ctx, r.x + 3, r.y + 1, inp.getText());

    if (!blink) {
        drawSolidBar(ctx, GfxRect(r.getRightX() - 10, r.getBottomY() - 5, 7, 2), COLOR_BLACK);
    }

    can.update(getExtent());
}

/****************************** UIPrefixArg ******************************/

/** Constructor. Make a prefix arg trigger widget. */
UIPrefixArg::UIPrefixArg()
{
    /* We don't want focus */
    setState(st_Disabled, true);
}

UIPrefixArg::~UIPrefixArg()
{ }

bool
UIPrefixArg::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        uint32 raw_key = event.key.code;
        if (!second_pass)
            raw_key &= ~ss_Alt;

        /* prefix arg cannot be 0, so we only accept 1..9 as initial keypress */
        if (raw_key >= '1' && raw_key <= '9') {
            startPrefixArg(raw_key - '0');
            return true;
        }
    }
    return UIFixedWidget::handleEvent(event, second_pass);
}

void
UIPrefixArg::drawContent(GfxCanvas& /*can*/)
{ }

/** Start entering a prefix argument. Opens the popup and lets the
    user enter the argument. When this function returns, the argument
    will have been handed to setPrefixArg() or posted to the event
    queue, if the user wanted one.
    \param n Initial value, should be nonzero */
void
startPrefixArg(int n)
{
    UIPrefixPopup popup(n);
    ui_root->add(popup);
    ui_root->run();
    ui_root->close(popup);
}
