/**
  *  \file ui/richdocument.h
  *  \brief Rich Document Display
  */
#ifndef PCC_V2_UI_RICHDOCUMENT_H
#define PCC_V2_UI_RICHDOCUMENT_H

#include <cstddef>
#include "ui/richtext.h"
#include "cpluslib/ptrvector.h"
#include "ui/ui.h"
#include "cpluslib/sigslot.h"

class GfxContext;

/** Rich Document. This class allows formatting of RichText. It breaks
    lines according to margins, and adjusts line heights according to
    used fonts.

    To use, configure the width and margin as needed, and add text
    using add(), addNewline(), etc., in any order you wish. This will
    accumulate all the text in an internal buffer of Item's
    (\c content); partially finished items will remain in an invisible
    intermediate buffer (\c last_chunk). To finish the document, call
    finish(). Afterwards, the RichDocument can be used for display. */
class RichDocument {
 public:
    struct Item {
        /* These values are sorted such that one can test '& Link' for link text */
        enum Kind {
            Normal     = 0,
            LinkTarget = 1,
            Link       = 0x80,
            LinkHover  = 0x81,
            LinkFocus  = 0x82
        };
        Kind kind;
        int x;
        int y;
        int w;           /* display width (0 for LinkTarget). */
        int font;
        UIColor::Color color;
        bool underline;
        string_t text;

        Item(Kind kind, int x, int y, int w, int font, UIColor::Color color, bool underline, const string_t& text)
            : kind(kind), x(x), y(y), w(w),
              font(font), color(color),
              underline(underline), text(text)
            { }
    };

    /** Link id. To users, this is an opaque value which uniquely identifies a link.
        Internally, it is the index into \c content pointing to the LinkTarget item. */
    typedef uint32_t linkid_t;
    static const linkid_t nil = linkid_t(-1);

    RichDocument();
    ~RichDocument();

    void clear();
    void setPageWidth(int width);
    void setLeftMargin(int lm);
    void setRightMargin(int rm);

    void add(const RichText& text);
    void add(const string_t& text);
    void add(const char* text);
    void addNewline();
    void addParagraph();
    void addAt(int x, const RichText& text);
    void addRight(int x, const RichText& text);
    void addCentered(int x, const RichText& text);
    void tabTo(int x);
    void finish();

    void draw(GfxContext& ctx, GfxRect area, int skipY);

    linkid_t getLinkFromPos(GfxPoint pt) const;
    string_t getLinkTarget(linkid_t link) const;
    void setLinkKind(linkid_t link, Item::Kind kind);

    linkid_t getNextLink(linkid_t id) const;
    linkid_t getNextLink(linkid_t id, GfxRect limit) const;
    linkid_t getPreviousLink(linkid_t id) const;
    linkid_t getPreviousLink(linkid_t id, GfxRect limit) const;
    bool isLinkVisible(linkid_t id, GfxRect limit) const;

 private:
    /* Up to first_this_line: completely rendered content.
       After first_this_line: x positions have been assigned, y positions have not been assigned yet */
    ptr_vector<Item> content;
    /* Contains a partial line with no positions assigned.
       If it contains spaces, it is guaranteed to fit into a
       line after the current x. */
    ptr_vector<Item> last_chunk;
    int x;
    int y;
    std::size_t first_this_line;

    int page_width;
    int left_margin;
    int right_margin;

    void process();
    void flushLine();
    void flushWord();
    void flushItems(std::size_t n);
};

/** Rich document widget. Displays a RichDocument and allows the user
    to interact with it. Since the RichDocument stores preformatted text,
    there are two ways to fill the document:
    - after doing layout, but before showing the widget, populate the
      document normally
    - while the widget is active, manipulate the document and finally
      call handleDocumentUpdate(). */
class UIRichDocument : public UIWidget {
 public:
    enum {
        fl_Tab    = 1,
        fl_Return = 2,
        fl_Space  = 4,

        /* Combination: single-page hypertext */
        fl_SingleHyper = fl_Tab + fl_Return + fl_Space
    };
    UIRichDocument(GfxPoint pref_size, uint16_t key_flags);

    RichDocument& getDocument();

    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
    void onResize(GfxRect oldSize, bool byLayout);
    bool handleEvent(const UIEvent& event, bool second_pass);
    void onStateChange(int state, bool enabled);
    void handleDocumentUpdate();

    Signal1<string_t> sig_link_click;

 private:
    void setSelectedLink(RichDocument::linkid_t link);
    void setHoverLink(RichDocument::linkid_t link);
    void setLink(RichDocument::linkid_t link);

    RichDocument doc;
    GfxPoint pref_size;

    uint16_t key_flags;
    bool mdown;
    RichDocument::linkid_t selected_link;
    RichDocument::linkid_t hover_link;
};

#endif
