/**
  *  \file ui/widgets/button.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets/button.h"
#include "gfx/palette.h"
#include "gfx/gfxcontext.h"
#include "ui/uiglobal.h"

/******************************** UIButton *******************************/

/** \class UIButton
    \brief Push button

    This class provides the look of a standard push-button. */

UIButton::~UIButton()
{ }

namespace {
    /* Definition of a button shape. A number consists of a number of components
       that have different count and color depending on the button's state.

                 aaaaaaaaaaaaaaaaaaaaaaaaaaaa    a,b  left/top lines
                 abbbbbbbbbbbbbbbbbbbbbbbbbbc    c    bottom/right line
                 ab.........................c    ...  body (with text)
                 ab.........................c
                 ab.........................c
                 ab.........................c
                 accccccccccccccccccccccccccc

       The shapes defined here are a 99% approximation of the "standard" shape
       used in PCC 1.x. 1.x had the fun config setting of a "button thickness",
       but I omit that here. The setting was a remainder of the original PCC 1.x
       button style which had an even thicker frame. */

    struct ButtonShape {
        uint8 num_left;         // Number of left/top lines
        uint8 left[4];          // Colors of left/top lines
        uint8 num_right;        // Number of right/bottom lines
        uint8 right[4];         // Colors of right/bottom lines
        uint8 body_color;       // Color of button body
    };
}

static const ButtonShape button_shapes[] = {
    // --- Normal Size ---
    // normal
    { 2, { COLOR_WHITE, COLOR_GRAYSCALE+10 }, 2, { COLOR_BLACK, COLOR_GRAYSCALE+8 }, COLOR_GRAYSCALE+9 },
    { 2, { COLOR_WHITE, COLOR_GRAYSCALE+11 }, 2, { COLOR_BLACK, COLOR_GRAYSCALE+9 }, COLOR_GRAYSCALE+10 },
    // normal highlight
    { 3, { COLOR_BLACK, COLOR_GRAYSCALE+6, COLOR_GRAYSCALE+9 },  2, { COLOR_WHITE, COLOR_GRAYSCALE+13 }, COLOR_GRAYSCALE+12 },
    { 3, { COLOR_BLACK, COLOR_GRAYSCALE+7, COLOR_GRAYSCALE+10 }, 2, { COLOR_WHITE, COLOR_GRAYSCALE+14 }, COLOR_GRAYSCALE+13 },
    // pressed
    { 3, { COLOR_BLACK, COLOR_GRAYSCALE+13, COLOR_GRAYSCALE+11 }, 2, { COLOR_GRAYSCALE+3, COLOR_GRAYSCALE+9 },  COLOR_GRAYSCALE+9 },
    { 3, { COLOR_BLACK, COLOR_GRAYSCALE+14, COLOR_GRAYSCALE+12 }, 2, { COLOR_GRAYSCALE+4, COLOR_GRAYSCALE+10 }, COLOR_GRAYSCALE+10 },
    // pressed highlight
    { 3, { COLOR_GRAYSCALE+0, COLOR_GRAYSCALE+5, COLOR_GRAYSCALE+7 }, 1, { COLOR_GRAYSCALE+15 }, COLOR_GRAYSCALE+12 },
    { 3, { COLOR_GRAYSCALE+1, COLOR_GRAYSCALE+6, COLOR_GRAYSCALE+8 }, 1, { COLOR_GRAYSCALE+15 }, COLOR_GRAYSCALE+13 },

    // --- Small Size ---
    // normal
    { 1, { COLOR_WHITE }, 1, { COLOR_BLACK }, COLOR_GRAYSCALE+9 },
    { 1, { COLOR_WHITE }, 1, { COLOR_BLACK }, COLOR_GRAYSCALE+10 },
    // normal highlight
    { 1, { COLOR_WHITE }, 1, { COLOR_BLACK }, COLOR_GRAYSCALE+12 },
    { 1, { COLOR_WHITE }, 1, { COLOR_BLACK }, COLOR_GRAYSCALE+13 },

    // pressed
    { 2, { COLOR_BLACK, COLOR_GRAYSCALE+9 },  1, { COLOR_WHITE }, COLOR_GRAYSCALE+9 },
    { 2, { COLOR_BLACK, COLOR_GRAYSCALE+9 },  1, { COLOR_WHITE }, COLOR_GRAYSCALE+10 },
    // pressed highlight
    { 2, { COLOR_BLACK, COLOR_GRAYSCALE+12 }, 1, { COLOR_WHITE }, COLOR_GRAYSCALE+12 },
    { 2, { COLOR_BLACK, COLOR_GRAYSCALE+12 }, 1, { COLOR_WHITE }, COLOR_GRAYSCALE+13 },
};

/** Draw a button. Exported to be callable from non-button widgets that contain
    button-lookalikes, such as scrollbars.
    \param can        Canvas to draw on
    \param extent     Dimensions of the button
    \param name       Text on button
    \param flags      Style flags (bf_XXX, st_XXX) */
void
UIButton::drawButton(GfxCanvas& can, GfxRect extent, const string_t& name, int flags)
{
    // convert flags into button_shapes index
    int slot = 0;
    if ((flags & bf_SmallSize) != 0)
        slot += 8;
    if ((flags & (bf_Pressed | bf_PressOverride)) != 0)
        slot += 4;
    if ((flags & bf_Highlight) != 0)
        slot += 2;
    if ((flags & st_Selected) != 0)
        slot += 1;

    const ButtonShape* sh = &button_shapes[slot];

    // draw button
    GfxRect e = extent;
    int delta = 0;
    for (int i = 0; i < 4; ++i) {
        if (i < sh->num_left) {
            // upper-left corner
            can.drawHLine(e.x, e.y, e.getRightX(),    standard_colors[sh->left[i]], SOLID_LINE, GFX_ALPHA_OPAQUE);
            can.drawVLine(e.x, e.y+1, e.getBottomY(), standard_colors[sh->left[i]], SOLID_LINE, GFX_ALPHA_OPAQUE);
            ++e.x;
            ++e.y;
            --e.w;
            --e.h;
            ++delta;
        }
        if (i < sh->num_right) {
            // lower-right corner
            can.drawHLine(e.x, e.getBottomY()-1, e.getRightX(),    standard_colors[sh->right[i]], SOLID_LINE, GFX_ALPHA_OPAQUE);
            can.drawVLine(e.getRightX()-1, e.y,  e.getBottomY()-1, standard_colors[sh->right[i]], SOLID_LINE, GFX_ALPHA_OPAQUE);
            --e.w;
            --e.h;
            --delta;
        }
    }
    can.drawBar(e, standard_colors[sh->body_color], GFX_TRANSPARENT, SOLID_FILL, GFX_ALPHA_OPAQUE);

    // draw text
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[(flags & bf_SmallSize) ? FONT_SMALL : FONT_TITLE]);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    
    if (flags & st_Disabled)
        ctx.setColor(COLOR_BLACK);
    else
        ctx.setColor(COLOR_SHIELD + 5);

    if (flags & bf_LeftJust) {
        ctx.setTextAlign(0, 1);
        outText(ctx, extent.x + 8 + delta, extent.y + extent.h/2 + delta, name);
    } else {
        ctx.setTextAlign(1, 1);
        outText(ctx, extent.x + extent.w/2 + delta, extent.y + extent.h/2 + delta, name);
    }
}

/** Draw the button. */
void
UIButton::drawContent(GfxCanvas& can)
{
    drawButton(can, extent, name, getState());
    can.update(extent);
}

void
UIButton::getLayoutInfo(LayoutInfo& info)
{
    if (hasState(bf_SmallSize)) {
        info.flags = 0;
        info.min_size = info.pref_size
            = GfxPoint(name.length() == 1 ? 15 : 8 + fonts[FONT_SMALL]->getTextWidth(name), 15);
    } else {
        info.flags = 0;
        info.min_size = info.pref_size
            = GfxPoint(name.length() == 1 ? 25 : 15 + fonts[FONT_TITLE]->getTextWidth(name), 25);
    }
}
