/**
  *  \file ui/widgets/decimalsel.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets/decimalsel.h"
#include "gfx/gfxcontext.h"
#include "ui/skin.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "util/misc.h"
#include "cpluslib/format.h"

/** Constructor.
    \param id Widget Id
    \param low Lower limit
    \param high Upper limit
    \param set Default step size */
UIDecimalSelector::UIDecimalSelector(int id, long low, long high, int step)
    : UINumberSelector(id, low, high, step),
      state(sNormal)
{ }

UIDecimalSelector::~UIDecimalSelector()
{ }

bool
UIDecimalSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (hasState(st_Focused) && e.type == ev_Key) {
        if (e.key.code >= '0' && e.key.code <= '9') {
            // Digit: type
            setSelectedWidget();
            long n = e.key.code - '0';
            if (state == sNormal)
                n += 10*getValue();
            if (getMin() <= n && n <= getMax()) {
                setValue(n);
                wState(sNormal);
            }
            return true;
        } else if (e.key.code == SDLK_BACKSPACE) {
            // Backspace: delete last digit
            setSelectedWidget();
            if (getValue() < 10)
                wState(sZeroed);
            else
                wState(sNormal);
            setValue(getValue() / 10);
            return true;
        } else if (e.key.code == SDLK_DELETE) {
            // Del: clear if selected
            setSelectedWidget();
            if (state == sTypeErase) {
                setValue(0);
                wState(sZeroed);
            }
            return true;
        } else if (e.key.code == ss_Ctrl + SDLK_DELETE || e.key.code == ss_Ctrl + 'y') {
            // Ctrl-Del, Ctrl-Y: clear
            setSelectedWidget();
            setValue(0);
            wState(sZeroed);
            return true;
        } else if (e.key.code == ss_Ctrl + 'u') {
            // Ctrl-U: unselect
            setSelectedWidget();
            wState(sNormal);
            return true;
        } else if (e.key.code == ss_Ctrl + 't') {
            // Ctrl-T: Swap last two digits
            // Because we cannot have leading zeroes, this will also turn 2 into 20 and vice versa.
            setSelectedWidget();
            wState(sNormal);
            long n = getValue();
            setValue(100*(n/100)
                     + 10*(n%10)
                     + (n/10)%10);
            return true;
        }
    }

    // FIXME: We must track whether the base class' logic changes our value,
    // and adapt our state accordingly. This is a hack, but it's easier than
    // the other way (hooking into sig_changed and ensure that we never trigger
    // ourselves). Plus, PCC 1.x does the same :-)
    long l = getValue();
    bool b = UINumberSelector::handleEvent(e, second_pass);
    if (state != sNormal && l != getValue())
        wState(sNormal);
    return b;
}

void
UIDecimalSelector::onStateChange(int state, bool enabled)
{
    if (state & st_Focused) {
        if (enabled)
            wState(sTypeErase);
        else
            wState(sNormal);
        drawWidget(false);
    }
    UINumberSelector::onStateChange(state, enabled);
}

void
UIDecimalSelector::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_TITLE]);

    string_t post, val;

    if (getFocusState() != NoFocus) {
        post = "_";              // FIXME: replace by proper cursor
    }

    if (hasState(pds_ShowMax))
        post += format(_(" (max. %d)"), getMax());

    if (state != sZeroed)
        val = getStringValue();

    int val_width  = ctx.getFont().getTextWidth(val);
    int post_width = ctx.getFont().getTextWidth(post);

    if (val_width >= getExtent().w)
        val_width = getExtent().w;
    if (post_width + val_width >= getExtent().w)
        post_width = getExtent().w - val_width;

    int remain = getExtent().w - post_width - val_width;

    /* Left-justify:  | val_width | post_width | suffix |
       Right-justify: | prefix | val_width | post_width | */
    
    int pre_width, suf_width;
    if (hasState(pds_RightJust)) {
        pre_width = remain;
        suf_width = 0;
    } else {
        pre_width = 0;
        suf_width = remain;
    }

    /* Draw it */
    const int x = getExtent().x, y = getExtent().y, h = getExtent().h;
    if (state == sTypeErase) {
        /* Focused and type-erase */
        drawBackground(ctx, GfxRect(x, y, pre_width, h));
        drawBackground(ctx, GfxRect(x + pre_width + val_width + post_width, y, suf_width, h));
        drawSolidBar(ctx, GfxRect(x + pre_width, y, val_width, h), UIColor::tc_Input);
        ctx.setColor(UIColor::tc_Background);
    } else {
        /* Not type-erase, thus everything on regular background */
        drawBackground(ctx, getExtent());
        ctx.setColor(getFocusState() != NoFocus ? UIColor::tc_Input : UIColor::tc_Static);
    }

    outTextF(ctx, x + pre_width, y, val_width, val);

    ctx.setColor(UIColor::tc_Static);
    outTextF(ctx, x + pre_width + val_width, y, post_width, post);

    can.update(getExtent());
}

void
UIDecimalSelector::wState(State st)
{
    if (st != state) {
        state = st;
        drawWidget(false);
    }
}

/** Get value as string. This should format the value of this widget
    (getValue()) as a string. Descendants can override this method to
    provide additional info. \default returns just the number */
string_t
UIDecimalSelector::getStringValue()
{
    return numToString(getValue());
}

void
UIDecimalSelector::getLayoutInfo(LayoutInfo& li)
{
    li.flags = li_FlexibleH;

    // FIXME? heuristic based on font_ems
    int width = (hasState(pds_ShowMax) ? font_ems[FONT_TITLE]*10 : font_ems[FONT_TITLE]*5);
    li.pref_size = li.min_size = GfxPoint(width, font_heights[FONT_TITLE]);
}
